﻿#define WIN32_LEAN_AND_MEAN
#define WIN32_EXTRA_LEAN
#include <Windows.h>

#define MAX_ITERATIONS 255
#define MAX_INPUT 8

////////////////////////////////////////////
// Necessary to avoid the runtime library //
////////////////////////////////////////////
void* __cdecl operator new(unsigned int cb)
{
  return HeapAlloc(GetProcessHeap(),0,cb);
}

void __cdecl operator delete(void* pv)
{
  if(pv)
    HeapFree(GetProcessHeap(),0,pv);
}

extern "C" int _cdecl _purecall(void)
{
  return 0;
}
////////////////////////////////////////////
// End of ugly necessities                //
////////////////////////////////////////////

int __cdecl main()
{
	// Get input and output handles
	HANDLE hConsoleOutput = GetStdHandle(STD_OUTPUT_HANDLE);
	HANDLE hConsoleInput = GetStdHandle(STD_INPUT_HANDLE);

	// Get size
	CONSOLE_SCREEN_BUFFER_INFO info;
	GetConsoleScreenBufferInfo(hConsoleOutput, &info);
	unsigned char Width = unsigned char(info.srWindow.Right + 1);
	unsigned char Height = unsigned char(info.srWindow.Bottom);

	// Set default values
	double step = 0.1;
	double xStart = 0;
	double yStart = 0;
	double limit = 4;

	// Used for input
	DWORD inputCount;
	INPUT_RECORD* input = new INPUT_RECORD[MAX_INPUT];

	// Used to redraw the screen
	COORD zero;
	zero.X = 0;
	char c = char(177);

	bool cont = true;

	COORD mousePosition, mousePrevious;
	mousePosition.X = 0;
	mousePosition.Y = 0;
	double mouseMovementX, mouseMovementY;
	bool mousePressed, mouseScrolled;
	float zoomDirection;

	SetConsoleTitleA("ConBrot");

	char* IntroText =
		"ConBrot - A text mode mandelbrot generator\n"
		"------------------------------------------\n"
		"By Scepheo\n"
		"\n"
		"Controls:\n"
		" - Click and drag to move\n"
		" - Scroll to zoom\n"
		" - ESC to quit\n"
		"\n"
		"Click to start...";

	// Intro
	WriteConsoleA(hConsoleOutput, IntroText, strlen(IntroText), NULL, NULL);
	ReadConsoleInputA(hConsoleInput, input, 0, &inputCount);

	while (cont)
	{
		// Reset values
		mousePrevious = mousePosition;
		mousePressed = false;
		mouseScrolled = false;

		// Get input
		DWORD inputCount;
		GetNumberOfConsoleInputEvents(hConsoleInput, &inputCount);
		if (inputCount > MAX_INPUT)
			inputCount = MAX_INPUT;
		if (inputCount)
		{
			ReadConsoleInputA(hConsoleInput, input, inputCount, &inputCount);
			for (DWORD i = 0; i < inputCount; i++)
			{
				if (input[i].EventType == MOUSE_EVENT)
				{
					mousePosition = input[i].Event.MouseEvent.dwMousePosition;
					mousePressed = (input[i].Event.MouseEvent.dwButtonState & FROM_LEFT_1ST_BUTTON_PRESSED) > 0;
					mouseScrolled = (input[i].Event.MouseEvent.dwEventFlags & MOUSE_WHEELED) > 0;

					if (mouseScrolled)
					{
						DWORD buttonState = input[i].Event.MouseEvent.dwButtonState;
						WORD highWord = buttonState >> 16;

						// This checks if the sign bit it set, i.e. < 0
						if (highWord & 0x80)
							zoomDirection = 1;
						else
							zoomDirection = -1;
					}
				}

				if (input[i].EventType == KEY_EVENT)
				{
					if (input[i].Event.KeyEvent.wVirtualKeyCode == VK_ESCAPE)
						cont = false;
				}
			}
		}

		// Update settings if needed
		if (mouseScrolled)
		{
			step += step * 0.1 * zoomDirection;
		}

		if (mousePressed)
		{
			mouseMovementX = mousePrevious.X - mousePosition.X;
			mouseMovementY = mousePrevious.Y - mousePosition.Y;

			xStart += step * mouseMovementX;
			yStart += step * mouseMovementY;
		}

		// Update our screen if needed
		if (mouseScrolled || mousePressed)
		{
			zero.Y = 0;

			for (double yPos = -Height / 2; yPos < Height / 2; yPos++)
			{
				SetConsoleCursorPosition(hConsoleOutput, zero);
				zero.Y++;

				double y = yStart + yPos * step * 2;
				for (double xPos = -Width / 2; xPos < Width / 2; xPos++)
				{
					double x = xStart + xPos * step;
					double xt2, yt2, xt = x, yt = y;
					unsigned char iterations = 0;

					// Calculate the mandelbrot value
					while ((xt2 = xt * xt) + (yt2 = yt * yt) < limit && iterations < MAX_ITERATIONS)
					{
						iterations++;
						yt = 2 * xt * yt + y;
						xt = xt2 - yt2 + x;
					}

					if (iterations == MAX_ITERATIONS)
						SetConsoleTextAttribute(hConsoleOutput, 0);
					else
						SetConsoleTextAttribute(hConsoleOutput, iterations);

					WriteConsoleA(hConsoleOutput, &c, 1, NULL, NULL);
				}
			}
		}

		Sleep(50);
	}

	// Quit
	return 0;
}